<?php

namespace Drupal\sitemap_status\Commands;

use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\sitemap_status\SitemapStatus;
use Drush\Commands\DrushCommands;

/**
 * Class SitemapStatusCommands.
 *
 * @package Drupal\sitemap_status\Commands
 */
class SitemapStatusCommands extends DrushCommands {

  use StringTranslationTrait;

  /**
   * @var \Drupal\sitemap_status\SitemapStatus
   */
  protected $sitemapStatus;

  /**
   * SitemapStatusCommands constructor.
   *
   * @param \Drupal\simple_sitemap\Simplesitemap $sitemap_status
   */
  public function __construct(SitemapStatus $sitemap_status) {
    $this->sitemapStatus = $sitemap_status;
  }

  /**
   * Checks the locations statuses from a sitemap.
   *
   * @command sitemap-status:check
   *
   * @usage drush sitemap-status:check
   *   Checks the sitemap locations http status.
   *
   * @usage drush sitemap-status:check --basic-auth=username:password
   *   Appends basic auth credentials to the locations status check.
   *   It is not uncommon to have pull request environments behind basic auth and
   *   sometimes, it also applies to the CLI.
   *
   * @usage drush sitemap-status:check --path-prefix=europe
   *   When a path prefix is being used. Example: a specific directory or a region based redirect
   *   to be able to distinguish other 301 redirects.
   *
   * @usage drush sitemap-status:check --async=true
   *   Fetches location status asynchronously, faster but uses more resources.
   *
   * @usage drush sitemap-status:check --concurrent-requests=10
   *   Maximum amount of concurrent requests while using async.
   *
   * @usage drush sitemap-status:check --excluded-urls=https://mysite.org,https://nginx-production.mysite.org
   *   Do not run the locations check for these environments.
   *   Without a trailing slash and with the scheme. Comma separated.
   *
   * @usage drush sitemap-status:check --print-headers=true
   *   Prints http headers.
   *
   * @validate-module-enabled sitemap_status
   *
   * @aliases ssc, sitemap-status-check
   */
  public function check(array $options = [
    'path-prefix' => '',
    'basic-auth' => '',
    'async' => '',
    'concurrent-requests' => '',
    'excluded-urls' => '',
    'print-headers' => '',
  ]) {
    // @todo dependency injection.
    $messenger = \Drupal::messenger();
    $currentHost = \Drupal::request()->getSchemeAndHttpHost();

    $excludedHosts = array_map('trim', array_filter(explode(',', $options['excluded-urls'])));
    // Possibly strip scheme.
    if (!empty($excludedHosts) && in_array($currentHost, $excludedHosts)) {
      $messenger->addMessage($this->t('The current host @current_host is excluded, skipping locations checks.', [
        '@current_host' => $currentHost,
      ]));
    }

    $basicAuth = '';
    if (!empty($options['basic-auth'])) {
      $basicAuth = $options['basic-auth'];
    }

    $pathPrefix = '';
    if (!empty($options['path-prefix'])) {
      $pathPrefix = $options['path-prefix'];
    }

    $async = FALSE;
    $concurrentRequests = NULL;
    if (!empty($options['async']) && $options['async'] === 'true') {
      $async = TRUE;
      if (!empty($options['concurrent-requests']) && is_numeric($options['concurrent-requests'])) {
        $concurrentRequests = (int) $options['concurrent-requests'];
      }
    }

    $printHeaders = FALSE;
    if (!empty($options['print-headers']) && $options['print-headers'] === 'true') {
      $printHeaders = TRUE;
    }

    $this->sitemapStatus->checkLocations(
      $basicAuth,
      $pathPrefix,
      $async,
      $concurrentRequests,
      $printHeaders
    );
  }

}
