<?php

namespace Drupal\sitemap_status\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\sitemap_status\SitemapStatusInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class SettingsForm.
 */
class SettingsForm extends ConfigFormBase {

  /**
   * Drupal\sitemap_status\SitemapStatusInterface definition.
   *
   * @var \Drupal\sitemap_status\SitemapStatusInterface
   */
  protected $sitemapStatus;

  /**
   * Drupal\Core\Queue\QueueFactory definition.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Drupal\Core\Datetime\DateFormatterInterface definition.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->sitemapStatus = $container->get('sitemap_status');
    $instance->queueFactory = $container->get('queue');
    $instance->dateFormatter = $container->get('date.formatter');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'sitemap_status.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'sitemap_status_settings';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('sitemap_status.settings');
    $summary = $this->sitemapStatus->getStatusSummary();
    /** @var \Drupal\Core\Render\RendererInterface $renderer */
    $renderer = \Drupal::service('renderer');
    $form['summary'] = [
      '#type' => 'details',
      '#title' => $this->t('Status summary'),
      '#open' => FALSE,
    ];
    $form['summary']['status_summary'] = [
      '#type' => 'item',
      '#markup' => $renderer->render($summary),
    ];
    $form['locations'] = [
      '#type' => 'details',
      '#title' => t('Locations'),
      '#open' => TRUE,
    ];
    $form['locations']['basic_auth'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Basic auth'),
      '#description' => $this->t('Appends basic auth credentials to the locations status check. Format: <em>username:password</em>'),
      '#maxlength' => 255,
      '#size' => 64,
      '#default_value' => $config->get('basic_auth'),
    ];
    $form['locations']['path_prefix'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Path prefix'),
      '#description' => $this->t('Can be used for a specific directory or a region based redirect, to be able to distinguish other 301 redirects. Without leading or trailing slashes.'),
      '#maxlength' => 64,
      '#size' => 64,
      '#default_value' => $config->get('path_prefix'),
    ];
    $form['locations']['async'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Async requests'),
      '#description' => $this->t('Fetches locations status asynchronously, faster but uses more resources.'),
      '#default_value' => $config->get('async'),
    ];
    $form['locations']['concurrent_requests'] = [
      '#type' => 'number',
      '#title' => $this->t('Maximum concurrent requests'),
      '#default_value' => $config->get('concurrent_requests') ?: SitemapStatusInterface::MAX_CONCURRENT_REQUESTS,
      '#min' => 2,
      '#states' => [
        'invisible' => [
          ':input[name="async"]' => ['checked' => FALSE],
        ],
      ],
    ];

    $form['trigger'] = [
      '#type' => 'details',
      '#title' => t('Trigger'),
      '#open' => TRUE,
    ];
    $form['trigger']['cron_frequency'] = [
      '#type' => 'select',
      '#title' => $this->t('Cron frequency'),
      '#default_value' => $config->get('cron_frequency'),
      '#options' => [
        0 => '<' . t('Disable cron check') . '>',
        1 => t('Daily'),
        7 => t('Weekly'),
      ],
      '#description' => $this->t('Optionally triggers the sitemap status on cron.'),
    ];

    $form['actions_message'] = [
      '#type' => 'item',
      '#markup' => $this->t('If the configuration has changed, it should be saved to apply while checking the status.'),
    ];
    $form = parent::buildForm($form, $form_state);
    $form['actions']['check_status'] = [
      '#type' => 'submit',
      '#value' => $this->t('Check http status'),
      '#submit' => [[$this, 'batchCheckStatus']],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);

    $this->config('sitemap_status.settings')
      ->set('production_url', $form_state->getValue('production_url'))
      ->set('path_prefix', $form_state->getValue('path_prefix'))
      ->set('basic_auth', $form_state->getValue('basic_auth'))
      ->set('async', $form_state->getValue('async'))
      ->set('concurrent_requests', $form_state->getValue('concurrent_requests'))
      ->set('cron_frequency', $form_state->getValue('cron_frequency'))
      ->save();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $pathPrefix = $form_state->getValue('path_prefix');
    $basicAuth = $form_state->getValue('basic_auth');
    $asyncRequests = $form_state->getValue('async');
    $concurrentRequests = (int) $form_state->getValue('concurrent_requests');

    // @todo improve validation / possibly refactor with sitemap status service.
    if (!empty($pathPrefix)) {
      $currentHost = \Drupal::request()->getSchemeAndHttpHost();
      $environmentUrl = $currentHost . '/' . $pathPrefix;
      if (!$this->sitemapStatus->isUrlValid($environmentUrl)) {
        $form_state->setError(
          $form['path_prefix'],
          $this->t('The resulting url with the path prefix (@environment_url) is not valid.', [
            '@environment_url' => $environmentUrl,
          ])
        );
      }
    }

    if (!empty($basicAuth)) {
      $basicAuthParts = explode(':', $basicAuth);
      if (count($basicAuthParts) !== 2) {
        $form_state->setError(
          $form['basic_auth'],
          $this->t('The basic auth must be with the format <em>username:password</em>.')
        );
      }
    }

    if ($asyncRequests === 1 && $concurrentRequests > SitemapStatusInterface::MAX_CONCURRENT_REQUESTS) {
      $this->messenger()->addWarning($this->t('The maximum amount of concurrent requests might vary depending on the available resources. It is advised to not configure more than @max_requests.', [
        '@max_requests' => SitemapStatusInterface::MAX_CONCURRENT_REQUESTS,
      ]));
    }
  }

  /**
   * Submit handler that checks the locations status with a batch.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function batchCheckStatus(array &$form, FormStateInterface $form_state) {
    $config = $this->config('sitemap_status.settings');
    $pathPrefix = $config->get('path_prefix');
    $basicAuth = $config->get('basic_auth');
    $asyncRequests = $config->get('async');

    if (!empty($basicAuth)) {
      $this->sitemapStatus->setBasicAuth($basicAuth);
    }
    if (!empty($pathPrefix)) {
      $this->sitemapStatus->setPathPrefix($pathPrefix);
    }

    $sitemapLocations = $this->sitemapStatus->fetchSitemapLocations();
    $operations = [];
    if ($asyncRequests === 1) {
      $concurrentRequests = $config->get('concurrent_requests');
      if (empty($concurrentRequests)) {
        $concurrentRequests = SitemapStatusInterface::MAX_CONCURRENT_REQUESTS;
      }
      $locationsGroups = array_chunk($sitemapLocations, $concurrentRequests);
      foreach ($locationsGroups as $locationGroup) {
        $operations[] = [
          '\Drupal\sitemap_status\LocationBatch::batchLocationsGroupProcess', [$locationGroup],
        ];
      }
    }
    else {
      foreach ($sitemapLocations as $location) {
        $operations[] = [
          '\Drupal\sitemap_status\LocationBatch::batchLocationProcess', [$location],
        ];
      }
    }

    $batch = [
      'title' => $this->t('Http status check of @total sitemap locations', ['@total' => count($sitemapLocations)]),
      'operations' => $operations,
      'finished' => ['\Drupal\sitemap_status\LocationBatch', 'batchFinished'],
    ];

    batch_set($batch);
  }

}
