<?php

namespace Drupal\sitemap_status;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\Core\Queue\SuspendQueueException;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;

/**
 * Class SitemapStatus.
 */
class SitemapStatus implements SitemapStatusInterface {

  use StringTranslationTrait;

  /**
   * Drupal\Core\Config\ConfigFactoryInterface definition.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Drupal\Core\Messenger\MessengerInterface definition.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Drupal\Core\Queue\QueueFactory definition.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Drupal\Core\Queue\QueueWorkerManagerInterface definition.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManagerInterface
   */
  protected $pluginManagerQueueWorker;

  /**
   * Drupal\Core\Cache\CacheBackendInterface definition.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cacheBackend;

  /**
   * @var string
   */
  private $basicAuth;

  /**
   * @var string
   */
  private $pathPrefix;

  /**
   * @var bool
   */
  private $printHeaders;

  /**
   * Constructs a new SitemapStatus object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    MessengerInterface $messenger,
    QueueFactory $queue_factory,
    QueueWorkerManagerInterface $plugin_manager_queue_worker,
    CacheBackendInterface $cache_backend
  ) {
    $this->configFactory = $config_factory;
    $this->messenger = $messenger;
    $this->queueFactory = $queue_factory;
    $this->pluginManagerQueueWorker = $plugin_manager_queue_worker;
    $this->cacheBackend = $cache_backend;
  }

  /**
   * {@inheritdoc}
   */
  public function checkLocations(
    $basic_auth = '',
    $path_prefix = '',
    $async = FALSE,
    $concurrent_requests = NULL,
    $print_headers = FALSE
  ) {
    $this->printHeaders = $print_headers;
    $this->setBasicAuth($basic_auth);
    $this->setPathPrefix($path_prefix);

    $sitemapLocations = $this->fetchSitemapLocations();

    if (!$async) {
      $locationsStatuses = $this->fetchLocationsByStatusSync($sitemapLocations);
    }
    else {
      $locationsStatuses = $this->fetchLocationsByStatusAsync($sitemapLocations, $concurrent_requests);
    }

    $groupedLocations = [];
    if (is_array($locationsStatuses)) {
      $groupedLocations = $this->groupLocationsByStatus($locationsStatuses);
      if (!empty($groupedLocations)) {
        \Drupal::logger('sitemap_status')->info($this->t('Sitemap status check has been completed.'));
      }
      else {
        \Drupal::logger('sitemap_status')->warning($this->t('Sitemap status check has not been completed.'));
      }
      // As a side effect, store in the cache.
      $this->cacheBackend->set(SitemapStatusInterface::CACHE_ID, $groupedLocations);
    }
    return $groupedLocations;
  }

  /**
   * {@inheritdoc}
   */
  public function setBasicAuth($basic_auth) {
    if (!empty($basic_auth)) {
      $basicAuthParts = explode(':', $basic_auth);
      if (count($basicAuthParts) !== 2) {
        $this->messenger->addError(
          $this->t('The basic auth must be with the format <em>username:password</em>. It has been ignored.')
        );
      }
      else {
        $this->basicAuth = $basic_auth;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setPathPrefix($path_prefix) {
    if (!empty($path_prefix)) {
      $currentHost = \Drupal::request()->getSchemeAndHttpHost();
      $environmentUrl = $currentHost . '/' . $path_prefix;
      if (!$this->isUrlValid($environmentUrl)) {
        $this->messenger->addError(
          $this->t('The resulting url with the path prefix (@environment_url) is not valid. It has been ignored.', [
            '@environment_url' => $environmentUrl,
          ])
        );
      }
      else {
        $this->pathPrefix = $path_prefix;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getLocationsStatuses() {
    $result = [];
    if (\Drupal::currentUser()->hasPermission('access sitemap status')) {
      if ($cache = $this->cacheBackend->get(SitemapStatusInterface::CACHE_ID)) {
        $result = $cache->data;
      }
      if (empty($result)) {
        // @todo add call to action to run the status check.
        $this->messenger->addWarning($this->t('No sitemap status yet.'));
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getLocationsStatusesByGroup() {
    $result = [];
    $locationsStatuses = $this->getLocationsStatuses();
    $statusGroups = [
      SitemapStatusInterface::STATUS_NOT_REACHED,
      SitemapStatusInterface::STATUS_2xx,
      SitemapStatusInterface::STATUS_3xx,
      SitemapStatusInterface::STATUS_4xx,
      SitemapStatusInterface::STATUS_5xx,
    ];
    foreach ($locationsStatuses as $httpsStatus => $locations) {
      foreach ($statusGroups as $httpStatusGroup) {
        if ($this->startsWith((string) $httpsStatus, $httpStatusGroup)) {
          if (!array_key_exists($httpStatusGroup, $result)) {
            $result[$httpStatusGroup] = [];
          }
          $result[$httpStatusGroup] += $locations;
        }
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatusLevel() {
    $statusGroups = $this->getLocationsStatusesByGroup();
    if (
      array_key_exists(SitemapStatusInterface::STATUS_5xx, $statusGroups) &&
      !empty($statusGroups[SitemapStatusInterface::STATUS_5xx])
    ) {
      return SitemapStatusInterface::REQUIREMENT_ERROR;
    }
    if (
      array_key_exists(SitemapStatusInterface::STATUS_4xx, $statusGroups) &&
      !empty($statusGroups[SitemapStatusInterface::STATUS_4xx])
    ) {
      return SitemapStatusInterface::REQUIREMENT_WARNING;
    }
    if (
      array_key_exists(SitemapStatusInterface::STATUS_NOT_REACHED, $statusGroups) &&
      !empty($statusGroups[SitemapStatusInterface::STATUS_NOT_REACHED])
    ) {
      return SitemapStatusInterface::REQUIREMENT_INFO;
    }
    if (
      array_key_exists(SitemapStatusInterface::STATUS_3xx, $statusGroups) &&
      !empty($statusGroups[SitemapStatusInterface::STATUS_3xx])
    ) {
      return SitemapStatusInterface::REQUIREMENT_OK;
    }
    if (
      array_key_exists(SitemapStatusInterface::STATUS_2xx, $statusGroups) &&
      !empty($statusGroups[SitemapStatusInterface::STATUS_2xx])
    ) {
      return SitemapStatusInterface::REQUIREMENT_OK;
    }
    return SitemapStatus::REQUIREMENT_UNKNOWN;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatusLevelIcon() {
    $result = '';
    $statusLevel = $this->getStatusLevel();
    switch ($statusLevel) {
      case SitemapStatus::REQUIREMENT_ERROR:
        $result = '❌';
        break;

      case SitemapStatus::REQUIREMENT_UNKNOWN:
        $result = '❓';
        break;

      case SitemapStatus::REQUIREMENT_WARNING:
        $result = '⚠️';
        break;

      case SitemapStatus::REQUIREMENT_INFO:
        $result = 'ℹ️';
        break;

      case SitemapStatus::REQUIREMENT_OK:
        $result = '✅';
        break;
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatusSummary() {
    if (empty($this->getLocationsStatuses())) {
      $build = [
        '#type' => 'inline_template',
        '#template' => '<p>{{ message }}</p>',
        '#context' => [
          // @todo add call to action to run the status check.
          'message' => $this->t('No sitemap status yet.'),
        ],
      ];
    }
    else {
      $groupedLocations = $this->getLocationsStatusesByGroup();
      $listItems = [];
      foreach ($groupedLocations as $locationGroup => $locations) {
        $listItems[] = $this->t('Locations with status @group_status: @locations_count', [
          '@group_status' => $locationGroup . 'xx',
          '@locations_count' => count($locations),
        ]);
      }
      $build = [
        'summary' => [
          '#type' => 'inline_template',
          '#template' => '<p>{{ message }}</p> {{ groups }}',
          '#context' => [
            'message' => $this->t('Sitemap status summary.'),
            'groups' => [
              '#theme' => 'item_list',
              '#list_type' => 'ul',
              '#items' => $listItems,
            ],
          ],
        ],
        'link' => [
          '#type' => 'link',
          '#title' => $this->t('Details'),
          '#url' => Url::fromRoute('sitemap_status.details'),
        ],
      ];
    }
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function isUrlValid($url) {
    // @todo delegate to Drupal url validator.
    $result = TRUE;
    if (!$url || !is_string($url) || !preg_match('/^http(s)?:\/\/[a-z0-9-]+(.[a-z0-9-]+)*(:[0-9]+)?(\/.*)?$/i', $url)) {
      $result = FALSE;
    }
    return $result;
  }

  /**
   * Checks if a string starts with another one.
   *
   * @param string $haystack
   * @param string $needle
   *
   * @return bool
   */
  private function startsWith($haystack, $needle) {
    $length = strlen($needle);
    return (substr($haystack, 0, $length) === $needle);
  }

  /**
   * Sets the default curl options.
   *
   * @param resource $curl_handle
   */
  private function setDefaultCurlOptions(&$curl_handle) {
    curl_setopt($curl_handle, CURLOPT_RETURNTRANSFER, TRUE);
    curl_setopt($curl_handle, CURLOPT_SSL_VERIFYHOST, FALSE);
    curl_setopt($curl_handle, CURLOPT_SSL_VERIFYPEER, FALSE);
    curl_setopt($curl_handle, CURLOPT_CONNECTTIMEOUT, 5);
    curl_setopt($curl_handle, CURLOPT_TIMEOUT, 20);
    if (!empty($this->basicAuth)) {
      curl_setopt($curl_handle, CURLOPT_USERPWD, $this->basicAuth);
    }
    // @todo add option to follow redirect
    // curl_setopt($curlHandle, CURLOPT_FOLLOWLOCATION, TRUE);.
  }

  /**
   * {@inheritdoc}
   */
  public function fetchSitemapLocations() {
    $result = [];
    $environmentUrl = \Drupal::request()->getSchemeAndHttpHost();
    if (!empty($this->pathPrefix)) {
      $environmentUrl .= '/' . $this->pathPrefix;
    }
    $simpleSitemapConfig = $this->configFactory->get('simple_sitemap.settings');
    $simpleSitemapBaseUrl = $simpleSitemapConfig->get('base_url');

    $curlHandle = curl_init();
    $this->setDefaultCurlOptions($curlHandle);
    curl_setopt($curlHandle, CURLOPT_URL, $environmentUrl . '/sitemap.xml');
    $sitemap = curl_exec($curlHandle);
    $httpStatus = curl_getinfo($curlHandle, CURLINFO_HTTP_CODE);
    curl_close($curlHandle);
    if ((int) $httpStatus === 200) {
      $xmlElement = new \SimpleXMLElement($sitemap);
      foreach ($xmlElement->url as $urlElement) {
        $location = (string) $urlElement->loc;
        // Replace the sitemap url for this environment if necessary:
        // if the environment url does not match the simple sitemap base url
        // or if a path prefix has been added.
        if ($simpleSitemapBaseUrl !== $environmentUrl) {
          $location = str_replace($simpleSitemapBaseUrl, $environmentUrl, $location);
        }
        $result[] = $location;
      }
    }
    else {
      $this->messenger->addError($this->t('The sitemap returned a status of @status. Check if it has already been generated (<code>drush ssg</code>) or if the environment is accessible.', [
        '@status' => $httpStatus,
      ]));
    }
    return $result;
  }

  /**
   * Fetches the sitemap locations status synchronously.
   *
   * @param string[] $sitemap_locations
   *
   * @return int[]
   *   Locations grouped by http status code.
   */
  private function fetchLocationsByStatusSync(array $sitemap_locations) {
    $result = [];
    $queueId = SitemapStatusInterface::LOCATION;
    /** @var \Drupal\Core\Queue\QueueInterface $queue */
    $queue = $this->queueFactory->get($queueId);
    // Remove previous queue items if any.
    $queue->deleteQueue();
    foreach ($sitemap_locations as $location) {
      $queue->createItem($location);
    }
    $result = $this->processQueue($queueId);
    return $result;
  }

  /**
   * Fetches the sitemap locations status asynchronously.
   *
   * @param string[] $sitemap_locations
   * @param int|null $concurrent_requests
   *
   * @return int[]
   *   Locations grouped by http status code.
   */
  private function fetchLocationsByStatusAsync(array $sitemap_locations, $concurrent_requests) {
    $result = [];
    $concurrentRequests = $concurrent_requests;
    if (empty($concurrentRequests)) {
      $concurrentRequests = SitemapStatusInterface::MAX_CONCURRENT_REQUESTS;
    }
    $locationsGroups = array_chunk($sitemap_locations, $concurrentRequests);
    $queueId = SitemapStatusInterface::LOCATIONS_GROUP;
    /** @var \Drupal\Core\Queue\QueueInterface $queue */
    $queue = $this->queueFactory->get($queueId);
    // Remove previous queue items if any.
    $queue->deleteQueue();
    foreach ($locationsGroups as $locationsGroup) {
      $queue->createItem($locationsGroup);
    }
    $result = $this->processQueue($queueId);
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function groupLocationsByStatus(array $locations) {
    $result = [];
    foreach ($locations as $location => $httpStatus) {
      if (!array_key_exists($httpStatus, $result)) {
        $result[$httpStatus] = [];
      }
      $result[$httpStatus][] = $location;
    }
    return $result;
  }

  /**
   * Processes queue items.
   *
   * @param string $queue_id
   *
   * @return array
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function processQueue($queue_id) {
    $result = [];
    /** @var \Drupal\Core\Queue\QueueInterface $queue */
    $queue = $this->queueFactory->get($queue_id);
    switch ($queue_id) {
      case SitemapStatusInterface::LOCATIONS_GROUP:
        /** @var \Drupal\Core\Queue\QueueWorkerInterface $queueWorker */
        $queueWorker = $this->pluginManagerQueueWorker->createInstance('locations_group');
        break;

      case SitemapStatusInterface::LOCATION:
        /** @var \Drupal\Core\Queue\QueueWorkerInterface $queueWorker */
        $queueWorker = $this->pluginManagerQueueWorker->createInstance('location');
        break;
    }
    while ($item = $queue->claimItem()) {
      try {
        $result += $queueWorker->processItem($item->data);
        $queue->deleteItem($item);
      }
      catch (SuspendQueueException $exception) {
        $this->messenger->addError($exception->getMessage());
        $queue->releaseItem($item);
        break;
      }
      catch (\Exception $exception) {
        $this->messenger->addError($exception->getMessage());
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function fetchLocationStatus($location) {
    $curlHandle = curl_init();
    $this->setDefaultCurlOptions($curlHandle);
    curl_setopt($curlHandle, CURLOPT_HEADER, TRUE);
    curl_setopt($curlHandle, CURLOPT_NOBODY, TRUE);
    curl_setopt($curlHandle, CURLOPT_URL, $location);
    $status = curl_exec($curlHandle);
    if ($status) {
      $httpStatus = curl_getinfo($curlHandle, CURLINFO_HTTP_CODE);
      $this->printLocationErrors($location, $httpStatus, $status);
    }
    else {
      $httpStatus = 0;
    }
    curl_close($curlHandle);
    return $httpStatus;
  }

  /**
   * {@inheritdoc}
   */
  public function fetchLocationStatusMultiple(array $locations) {
    $result = [];
    $locationsCount = count($locations);
    $curlHandles = [];
    $curlMultipleHandle = curl_multi_init();
    for ($i = 0; $i < $locationsCount; $i++) {
      $location = $locations[$i];
      $curlHandles[$i] = curl_init($location);
      $this->setDefaultCurlOptions($curlHandles[$i]);
      curl_setopt($curlHandles[$i], CURLOPT_HEADER, TRUE);
      curl_setopt($curlHandles[$i], CURLOPT_NOBODY, TRUE);
      curl_multi_add_handle($curlMultipleHandle, $curlHandles[$i]);
    }

    do {
      $status = curl_multi_exec($curlMultipleHandle, $running);
      if ($running) {
        curl_multi_select($curlMultipleHandle);
      }
    } while ($running && $status === CURLM_OK);

    for ($i = 0; $i < $locationsCount; $i++) {
      $httpStatus = curl_getinfo($curlHandles[$i], CURLINFO_HTTP_CODE);
      // @todo make sure to refer to the right location.
      $result[$locations[$i]] = $httpStatus;
      // @todo implement specific case for curl status if we want to print headers.
      $this->printLocationErrors($locations[$i], $httpStatus);
      curl_multi_remove_handle($curlMultipleHandle, $curlHandles[$i]);
    }
    curl_multi_close($curlMultipleHandle);
    return $result;
  }

  /**
   * Prints a curl_exec result for a location.
   *
   * @param $location
   * @param $http_status
   * @param $curl_status
   */
  private function printLocationErrors($location, $http_status, $curl_status = NULL) {
    if ($http_status === 0 || $http_status >= 400) {
      $this->messenger->addError($this->t('Location @location: status @status.', [
        '@location' => $location,
        '@status' => $http_status,
      ]));
      if ($this->printHeaders && !is_null($curl_status)) {
        $this->messenger->addMessage(print_r($curl_status));
      }
    }
  }

}
